from abc import ABCMeta, abstractmethod
from typing import Optional, Iterable
from ontobio.model.similarity import SimResult


class SimilarityEngine(metaclass=ABCMeta):
    """
    Interface for similarity engines, methods for search and compare

    Differs from ontobio.sim.api.interfaces.SimApi in that a
    similarity engine is type specific, eg phenotype, go term
    and handles individuals appropriately
    """

    @abstractmethod
    def compare(self,
                reference_ids: Iterable,
                query_profiles: Iterable[Iterable],
                method: Optional) -> SimResult:
        """
        Given two lists of entities (classes, individuals),
        resolves them to some type (phenotypes, go terms, etc) and
        returns their similarity
        """
        pass

    @abstractmethod
    def search(self,
               id_list: Iterable,
               negated_ids: Iterable,
               limit: Optional[int],
               taxon_filter: Optional,
               category_filter: Optional,
               method: Optional)-> SimResult:
        """
        Given an input iterable of classes or individuals,
        resolves to target classes (phenotypes, go terms, etc)
        and provides a ranking of similar profiles
        """
        pass



import json
import typing
import collections
import enum
import datetime

from ontobio.ecomap import EcoMap
ecomap = EcoMap()
ecomap.mappings()


from typing import List, Optional, NamedTuple, Dict, Callable
from dataclasses import dataclass

Aspect = typing.NewType("Aspect", str)
Curie = typing.NewType("Curie", str)
Provider = typing.NewType("Provider", str)
Date = typing.NewType("Date", str)

@dataclass
class Subject:
    id: Curie
    label: str
    fullname: str
    synonyms: List[str]
    type: str
    taxon: Curie

@dataclass
class Term:
    id: Curie
    taxon: Curie

@dataclass
class Evidence:
    type: Curie # Curie of the ECO class
    has_supporting_reference: List[Curie]
    with_support_from: List[Curie]

@dataclass(unsafe_hash=True)
class ExtensionUnit:
    relation: Curie
    term: Curie

    def __str__(self) -> str:
        return "{relation}({term})".format(relation=self.relation, term=self.term)

@dataclass(unsafe_hash=True)
class ExtensionConjunctions:
    extensions: List[ExtensionUnit]

    def __str__(self) -> str:
        return ",".join([str(conj) for conj in self.extensions])

@dataclass
class ExtensionExpression:
    """
    ExtensionExpression ::= ConjunctionExpression { "|" ConjunctionExpression }
    ConjunctionExpression ::= ExtensionUnit { "," ExtensionUnit }
    ExtensionUnit ::= Relation "(" Term ")"
    """
    conjunctions: List[ExtensionConjunctions]

    def __str__(self) -> str:
        return "|".join([str(conjunction) for conjunction in self.conjunctions])

@dataclass(repr=True, unsafe_hash=True)
class GoAssociation:
    source_line: Optional[str]
    subject: Subject
    relation: Curie # This is the relation Curie
    object: Term
    negated: bool
    qualifiers: List[Curie]
    aspect: Optional[Aspect]
    interacting_taxon: Optional[Curie]
    evidence: Evidence
    subject_extensions: List[ExtensionUnit]
    object_extensions: ExtensionExpression
    provided_by: Provider
    date: Date
    properties: Dict[Curie, List[str]]

    def to_gaf_tsv(self) -> List:
        gp_isoforms = "" if not self.subject_extensions else self.subject_extensions[0].term
        db, subid = self.subject.id.split(":", maxsplit=1)
        qualifier = "|".join(self.qualifiers)
        if self.negated:
            qualifier = "NOT|{}".format(qualifier)

        taxon = self.object.taxon.replace("NCBITaxon", "taxon")
        if self.interacting_taxon:
            taxon = "{taxon}|{interacting}".format(taxon=taxon, interacting=self.interacting_taxon)

        return [
            db,
            subid,
            self.subject.label,
            qualifier,
            self.object.id,
            "|".join(self.evidence.has_supporting_reference),
            ecomap.ecoclass_to_coderef(self.evidence.type)[0],
            "|".join(self.evidence.with_support_from),
            self.aspect if self.aspect else "",
            self.subject.fullname,
            "|".join(self.subject.synonyms),
            self.subject.type,
            taxon,
            self.date,
            self.provided_by,
            str(self.object_extensions),
            gp_isoforms
        ]

    def to_gpad_tsv(self) -> List:
        db, subid = self.subject.id.split(":", maxsplit=1)
        qualifiers = "|".join(self.qualifiers)
        if self.negated:
            qualifiers = "NOT|{}".format(qualifiers)

        props_list = ["{key}={value}".format(key=key, value=value) for key, value in self.properties.items()]
        return [
            db,
            subid,
            qualifiers,
            self.object.id,
            "|".join(self.evidence.has_supporting_reference),
            self.evidence.type,
            "|".join(self.evidence.with_support_from),
            self.interacting_taxon if self.interacting_taxon else "",
            self.date,
            self.provided_by,
            str(self.object_extensions),
            "|".join(props_list)
        ]



"""Lexical mapping of ontology classes

The core data structure used here is a Mapping Graph. This is a
networkx Graph object (i.e. singly labeled, non-directional) that
connects lexically mapped nodes between two ontologies.

Edge Properties
---------------

idpair: (string,string)
    the pair of identifiers mapped
score: number
    Number between 0 and 100 indicating strength of match based on multiple criteria
synonyms: (Synonym,Synonym)
    pair of Synonym objects (including primary labels) used to create mapping
simscores: (number, number)
    Semantic similarity A to B and B to A respectively.
    Note that false positives or negatives in the ancestors or descendants in the xref graph will lead to bias in these scores.
reciprocal_score: int
    A number between 0 and 4 that indicates whether this was a reciprocal best match (RBM), with additional gradation based on whether
    ties are included. We distinguish between a true BM and a tied BM. 4 indicates true RBM. 1 indicates reciprocal tied BM (ie both are tied BMs). 2 indicates a combo of
    a true BM and a tied BM.
    Note that ties are less likely if semantic similarity is considered in the match.

"""
import networkx as nx
from networkx.algorithms import strongly_connected_components
import logging
import re
from ontobio.ontol import Synonym, Ontology
from collections import defaultdict
import pandas as pd
import numpy as np
import math

from marshmallow import Schema, fields, pprint, post_load

LABEL_OR_EXACT = 'label_or_exact'
logger = logging.getLogger(__name__)


def logit(p):
    return math.log2(p/(1-p))
def inv_logit(w):
    return 1/(1+2**(-w))

def default_wsmap():
    """
    Default word to normalized synonym list
    """
    return {
        'a':'',
        'of':'',
        'the':'',
        'i':'1',
        'ii':'2',
        'iii':'3',
        'iv':'4',
        'v':'5',
        'vi':'6',
        'vii':'7',
        'viii':'8',
        'ix':'9',
        'x':'10',
        'xi':'11',
        'xii':'12',
        'xiii':'13',
        'xiv':'14',
        'xv':'15',
        'xvi':'16',
        'xvii':'17',
        'xviii':'18',
        'xix':'19',
        'xx':'20',
        '':''
    }

class LexicalMapEngine():
    """
    generates lexical matches between pairs of ontology classes
    """

    SCORE='score'
    LEXSCORE='lexscore'
    SIMSCORES='simscores'
    CONDITIONAL_PR='cpr'
    
    def __init__(self, wsmap=default_wsmap(), config=None):
        """
        Arguments
        ---------
        wdmap: dict
            maps words to normalized synonyms.
        config: dict
            A configuration conforming to LexicalMapConfigSchema
        """
        # maps label or syn value to Synonym object
        self.lmap = {}
        # maps node id to synonym objects
        self.smap = {}
        self.wsmap = wsmap
        self.npattern = re.compile('[\W_]+')
        self.exclude_obsolete = True
        self.ontology_pairs = None
        self.id_to_ontology_map = defaultdict(list)
        self.merged_ontology = Ontology()
        self.config = config if config is not None else {}
        self.stats = {}

    def index_ontologies(self, onts):
        logger.info('Indexing: {}'.format(onts))
        for ont in onts:
            self.index_ontology(ont)
        
    def index_ontology(self, ont):
        """
        Adds an ontology to the index

        This iterates through all labels and synonyms in the ontology, creating an index
        """
        self.merged_ontology.merge([ont])
        syns = ont.all_synonyms(include_label=True)
        
        include_id = self._is_meaningful_ids()
        logger.info("Include IDs as synonyms: {}".format(include_id))
        if include_id:
            for n in ont.nodes():
                v = n
                # Get fragment
                if v.startswith('http'):
                    v = re.sub('.*/','',v)
                    v = re.sub('.*#','',v)
                syns.append(Synonym(n, val=v, pred='label'))
        
        logger.info("Indexing {} syns in {}".format(len(syns),ont))
        logger.info("Distinct lexical values: {}".format(len(self.lmap.keys())))
        for syn in syns:
            self.index_synonym(syn, ont)
        for nid in ont.nodes():
            self.id_to_ontology_map[nid].append(ont)

    def label(self, nid):
        return self.merged_ontology.label(nid)
    
    def index_synonym(self, syn, ont):
        """
        Index a synonym

        Typically not called from outside this object; called by `index_ontology`
        """
        if not syn.val:
            if syn.pred == 'label':
                if not self._is_meaningful_ids():
                    if not ont.is_obsolete(syn.class_id):
                        pass
                        #logger.error('Use meaningful ids if label not present: {}'.format(syn))
            else:
                logger.warning("Incomplete syn: {}".format(syn))
            return
        if self.exclude_obsolete and ont.is_obsolete(syn.class_id):
            return

        syn.ontology = ont
        prefix,_ = ont.prefix_fragment(syn.class_id)
        
        v = syn.val

        caps_match = re.match('[A-Z]+',v)
        if caps_match:
            # if > 75% of length is caps, assume abbreviation
            if caps_match.span()[1] >= len(v)/3:
                syn.is_abbreviation(True)
                
        # chebi 'synonyms' are often not real synonyms
        # https://github.com/ebi-chebi/ChEBI/issues/3294
        if not re.match('.*[a-zA-Z]',v):
            if prefix != 'CHEBI':
                logger.warning('Ignoring suspicous synonym: {}'.format(syn))
            return
        
        v = self._standardize_label(v)

        # TODO: do this once ahead of time
        wsmap = {}
        for w,s in self.wsmap.items():
            wsmap[w] = s
        for ss in self._get_config_val(prefix,'synsets',[]):
            # TODO: weights
            wsmap[ss['synonym']] = ss['word']
        nv = self._normalize_label(v, wsmap)
        
        self._index_synonym_val(syn, v)
        nweight = self._get_config_val(prefix, 'normalized_form_confidence', 0.8)
        if nweight > 0 and not syn.is_abbreviation():
            if nv != v:
                nsyn = Synonym(syn.class_id,
                               val=syn.val,
                               pred=syn.pred,
                               lextype=syn.lextype,
                               ontology=ont,
                               confidence=syn.confidence * nweight)
                self._index_synonym_val(nsyn, nv)

    def _index_synonym_val(self, syn, v):
        lmap = self.lmap
        smap = self.smap
        cid = syn.class_id
        if v not in lmap:
            lmap[v] = []
        lmap[v].append(syn)
        if cid not in smap:
            smap[cid] = []
        smap[cid].append(syn)

    def _standardize_label(self, v):
        # Add spaces separating camelcased strings
        v = re.sub('([a-z])([A-Z])',r'\1 \2',v)
        
        # always use lowercase when comparing
        # we may want to make this configurable in future
        v = v.lower()
        return v

    def _normalize_label(self, s, wsmap):
        """
        normalized form of a synonym
        """
        toks = []
        for tok in list(set(self.npattern.sub(' ', s).split(' '))):
            if tok in wsmap:
                tok=wsmap[tok]
            if tok != "":
                toks.append(tok)
        toks.sort()
        return " ".join(toks)

    def _get_config_val(self, prefix, k, default=None):
        v = None
        for oc in self.config.get('ontology_configurations', []):
            if prefix == oc.get('prefix', ''):
                v = oc.get(k, None)
        if v is None:
            v = self.config.get(k, None)
        if v is None:
            v = default
        return v
    
    def _is_meaningful_ids(self):
        return self.config.get('meaningful_ids', False)
    
    def find_equiv_sets(self):
        return self.lmap

    def get_xref_graph(self):
        """

        Generate mappings based on lexical properties and return as nx graph.

        Algorithm
        ~~~~~~~~~

        - A dictionary is stored between ref:`Synonym` values and synonyms. See ref:`index_synonym`.
          Note that Synonyms include the primary label

        - Each key in the dictionary is examined to determine if there exist two Synonyms from
          different ontology classes

        This avoids N^2 pairwise comparisons: instead the time taken is linear

        After initial mapping is made, additional scoring is performed on each mapping

        Edge properties
        ~~~~~~~~~~~~~~~
        The return object is a nx graph, connecting pairs of ontology classes.

        Edges are annotated with metadata about how the match was found:

        syns: pair
            pair of `Synonym` objects, corresponding to the synonyms for the two nodes
        score: int
            score indicating strength of mapping, between 0 and 100

        Returns
        -------
        Graph
            nx graph (bidirectional)
        """

        # initial graph; all matches
        g = nx.MultiDiGraph()

        # lmap collects all syns by token
        items = self.lmap.items()
        logger.info("collecting initial xref graph, items={}".format(len(items)))
        i = 0
        sum_nsyns = 0
        n_skipped = 0
        has_self_comparison = False
        if self.ontology_pairs:
            for (o1id,o2id) in self.ontology_pairs:
                if o1id == o2id:
                    has_self_comparison = True

        for (v,syns) in items:
            sum_nsyns += len(syns)
            i += 1
            if i % 1000 == 1:
                logger.info('{}/{}  lexical items avgSyns={}, skipped={}'.format(i,len(items), sum_nsyns/len(items), n_skipped))
            if len(syns) < 2:
                n_skipped += 1
                next
            if len(syns) > 10:
                logger.info('Syns for {} = {}'.format(v,len(syns)))
            for s1 in syns:
                s1oid = s1.ontology.id
                s1cid = s1.class_id
                for s2 in syns:
                    # optimization step: although this is redundant with _is_comparable,
                    # we avoid inefficient additional calls
                    if s1oid == s2.ontology.id and not has_self_comparison:
                        next
                    if s1cid != s2.class_id:
                        if self._is_comparable(s1,s2):
                            g.add_edge(s1.class_id, s2.class_id, syns=(s1,s2))

        logger.info("getting best supporting synonym pair for each match")
        # graph of best matches
        xg = nx.Graph()
        for i in g.nodes():
            for j in g.neighbors(i):
                best = 0
                bestm = None
                for m in g.get_edge_data(i,j).values():
                    (s1,s2) = m['syns']
                    score = self._combine_syns(s1,s2)
                    if score > best:
                        best = score
                        bestm = m
                syns = bestm['syns']
                xg.add_edge(i, j,
                            score=best,
                            lexscore=best,
                            syns=syns,
                            idpair=(i,j))

        self.score_xrefs_by_semsim(xg)
        self.assign_best_matches(xg)
        if self.merged_ontology.xref_graph is not None:
            self.compare_to_xrefs(xg, self.merged_ontology.xref_graph)
        else:
            logger.error("No xref graph for merged ontology")
        logger.info("finished xref graph")
        return xg

    # true if syns s1 and s2 should be compared.
    #  - if ontology_pairs is set, then only consider (s1,s2) if their respective source ontologies are in the list of pairs
    #  - otherwise compare all classes, but only in one direction
    def _is_comparable(self, s1, s2):
        if s1.class_id == s2.class_id:
            return False
        if self.ontology_pairs is not None:
            #logger.debug('TEST: {}{} in {}'.format(s1.ontology.id, s2.ontology.id, self.ontology_pairs))
            return (s1.ontology.id, s2.ontology.id) in self.ontology_pairs
        else:
            return s1.class_id < s2.class_id

    def _blanket(self, nid):
        nodes = set()
        for ont in self.id_to_ontology_map[nid]:
            nodes.update(ont.ancestors(nid))
            nodes.update(ont.descendants(nid))
        return list(nodes)
    
    def score_xrefs_by_semsim(self, xg, ont=None):
        """
        Given an xref graph (see ref:`get_xref_graph`), this will adjust scores based on
        the semantic similarity of matches.
        """
        logger.info("scoring xrefs by semantic similarity for {} nodes in {}".format(len(xg.nodes()), ont))
        for (i,j,d) in xg.edges(data=True):
            pfx1 = self._id_to_ontology(i)
            pfx2 = self._id_to_ontology(j)
            ancs1 = self._blanket(i)
            ancs2 = self._blanket(j)
            s1,_,_ = self._sim(xg, ancs1, ancs2, pfx1, pfx2)
            s2,_,_ = self._sim(xg, ancs2, ancs1, pfx2, pfx1)
            s = 1 - ((1-s1) * (1-s2))
            logger.debug("Score {} x {} = {} x {} = {} // {}".format(i,j,s1,s2,s, d))
            xg[i][j][self.SIMSCORES] = (s1,s2)
            xg[i][j][self.SCORE] *= s

    def _sim(self, xg, ancs1, ancs2, pfx1, pfx2):
        """
        Compare two lineages
        """
        xancs1 = set()
        for a in ancs1:
            if a in xg:
                # TODO: restrict this to neighbors in single ontology
                for n in xg.neighbors(a):
                    pfx = self._id_to_ontology(n)
                    if pfx == pfx2:
                        xancs1.add(n)
        logger.debug('SIM={}/{} ## {}'.format(len(xancs1.intersection(ancs2)), len(xancs1), xancs1.intersection(ancs2), xancs1))
        n_shared = len(xancs1.intersection(ancs2))
        n_total = len(xancs1)
        return (1+n_shared) / (1+n_total), n_shared, n_total

    # given an ontology class id,
    # return map keyed by ontology id, value is a list of (score, ext_class_id) pairs
    def _neighborscores_by_ontology(self, xg, nid):
        xrefmap = defaultdict(list)
        for x in xg.neighbors(nid):
            score = xg[nid][x][self.SCORE]
            for ont in self.id_to_ontology_map[x]:
                xrefmap[ont.id].append( (score,x) )
        return xrefmap
    
    # normalize direction
    def _dirn(self, edge, i, j):
        if edge['idpair'] == (i,j):
            return 'fwd'
        elif edge['idpair'] == (j,i):
            return 'rev'
        else:
            return None

    def _id_to_ontology(self, id):
        return self.merged_ontology.prefix(id)
        #onts = self.id_to_ontology_map[id]
        #if len(onts) > 1:
        #    logger.warning(">1 ontology for {}".format(id))
        
    def compare_to_xrefs(self, xg1, xg2):
        """
        Compares a base xref graph with another one
        """
        ont = self.merged_ontology
        for (i,j,d) in xg1.edges(data=True):
            ont_left = self._id_to_ontology(i)
            ont_right = self._id_to_ontology(j)
            unique_lr = True
            num_xrefs_left = 0
            same_left = False
            if i in xg2:
                for j2 in xg2.neighbors(i):
                    ont_right2 = self._id_to_ontology(j2)
                    if ont_right2 == ont_right:
                        unique_lr = False
                        num_xrefs_left += 1
                        if j2 == j:
                            same_left = True
            unique_rl = True
            num_xrefs_right = 0
            same_right = False
            if j in xg2:
                for i2 in xg2.neighbors(j):
                    ont_left2 = self._id_to_ontology(i2)
                    if ont_left2 == ont_left:
                        unique_rl = False
                        num_xrefs_right += 1
                        if i2 == i:
                            same_right = True

            (x,y) = d['idpair']
            xg1[x][y]['left_novel'] = num_xrefs_left==0
            xg1[x][y]['right_novel'] = num_xrefs_right==0
            xg1[x][y]['left_consistent'] = same_left
            xg1[x][y]['right_consistent'] = same_right

        
    
    def assign_best_matches(self, xg):
        """
        For each node in the xref graph, tag best match edges
        """
        logger.info("assigning best matches for {} nodes".format(len(xg.nodes())))
        for i in xg.nodes():
            xrefmap = self._neighborscores_by_ontology(xg, i)
            for (ontid,score_node_pairs) in xrefmap.items():
                score_node_pairs.sort(reverse=True)
                (best_score,best_node) = score_node_pairs[0]
                logger.info("BEST for {}: {} in {} from {}".format(i, best_node, ontid, score_node_pairs))
                edge = xg[i][best_node]
                dirn = self._dirn(edge, i, best_node)
                best_kwd = 'best_' + dirn
                if len(score_node_pairs) == 1 or score_node_pairs[0] > score_node_pairs[1]:
                    edge[best_kwd] = 2
                else:
                    edge[best_kwd] = 1
                for (score,j) in score_node_pairs:
                    edge_ij = xg[i][j]
                    dirn_ij = self._dirn(edge_ij, i, j)
                    edge_ij['cpr_'+dirn_ij] = score / sum([s for s,_ in score_node_pairs])
        for (i,j,edge) in xg.edges(data=True):
            # reciprocal score is set if (A) i is best for j, and (B) j is best for i
            rs = 0
            if 'best_fwd' in edge and 'best_rev' in edge:
                rs = edge['best_fwd'] * edge['best_rev']
            edge['reciprocal_score'] = rs
            edge['cpr'] = edge['cpr_fwd'] * edge['cpr_rev']

    def _best_match_syn(self, sx, sys, scope_map):
        """
        The best match is determined by the highest magnitude weight
        """
        SUBSTRING_WEIGHT = 0.2
        WBEST = None
        sbest = None
        sxv = self._standardize_label(sx.val)
        sxp = self._id_to_ontology(sx.class_id)
        for sy in sys:
            syv = self._standardize_label(sy.val)
            syp = self._id_to_ontology(sy.class_id)
            W = None
            if sxv == syv:
                confidence = sx.confidence * sy.confidence
                if sx.is_abbreviation() or sy.is_abbreviation:
                    confidence *= self._get_config_val(sxp, 'abbreviation_confidence', 0.5)
                    confidence *= self._get_config_val(syp, 'abbreviation_confidence', 0.5)
                W = scope_map[sx.scope()][sy.scope()] + logit(confidence/2)
            elif sxv in syv:
                W = np.array((-SUBSTRING_WEIGHT, SUBSTRING_WEIGHT, 0, 0))
            elif syv in sxv:
                W = np.array((SUBSTRING_WEIGHT, -SUBSTRING_WEIGHT, 0, 0))
            if W is not None:
                # The best match is determined by the highest magnitude weight
                if WBEST is None or max(abs(W)) > max(abs(WBEST)):
                    WBEST = W
                    sbest = sy
        return WBEST, sbest
    
    def weighted_axioms(self, x, y, xg):
        """
        return a tuple (sub,sup,equiv,other) indicating estimated prior probabilities for an interpretation of a mapping
        between x and y.

        See kboom paper
        """
        # TODO: allow additional weighting
        # weights are log odds w=log(p/(1-p))
        # (Sub,Sup,Eq,Other)
        scope_pairs = [
            ('label',   'label',   0.0, 0.0, 3.0,-0.8), 
            ('label',   'exact',   0.0, 0.0, 2.5,-0.5), 
            ('label',   'broad',  -1.0, 1.0, 0.0, 0.0), 
            ('label',   'narrow',  1.0,-1.0, 0.0, 0.0), 
            ('label',   'related', 0.0, 0.0, 0.0, 0.0), 
            ('exact',   'exact',   0.0, 0.0, 2.5,-0.5), 
            ('exact',   'broad',  -1.0, 1.0, 0.0, 0.0), 
            ('exact',   'narrow',  1.0,-1.0, 0.0, 0.0), 
            ('exact',   'related', 0.0, 0.0, 0.0, 0.0), 
            ('related', 'broad',  -0.5, 0.5, 0.0, 0.0), 
            ('related', 'narrow',  0.5,-0.5, 0.0, 0.0), 
            ('related', 'related', 0.0, 0.0, 0.0, 0.0), 
            ('broad',   'broad',   0.0, 0.0, 0.0, 1.0), 
            ('broad',   'narrow', -0.5, 0.5, 0.0, 0.2), 
            ('narrow',  'narrow',  0.0, 0.0, 0.0, 0.0)
        ]
        # populate symmetric lookup matrix
        scope_map = defaultdict(dict)
        for (l,r,w1,w2,w3,w4) in scope_pairs:
            l = l.upper()
            r = r.upper()
            scope_map[l][r] = np.array((w1,w2,w3,w4))
            scope_map[r][l] = np.array((w2,w1,w3,w4))

        # TODO: get prior based on ontology pair
        # cumulative sum of weights
        WS = None
        pfx1 = self._id_to_ontology(x)
        pfx2 = self._id_to_ontology(y)
        for mw in self.config.get('match_weights', []):
            mpfx1 = mw.get('prefix1','')
            mpfx2 = mw.get('prefix2','')
            X = np.array(mw['weights'])
            if mpfx1 == pfx1 and mpfx2 == pfx2:
                WS = X
            elif mpfx2 == pfx1 and mpfx1 == pfx2:
                WS = self._flipweights(X)
            elif mpfx1 == pfx1 and mpfx2 == '' and WS is None:
                WS = X
            elif mpfx2 == pfx1 and mpfx1 == '' and WS is None:
                WS = self._flipweights(X)

        if WS is None:
            WS = np.array((0.0, 0.0, 0.0, 0.0))
        # defaults
        WS += np.array(self.config.get('default_weights', [0.0, 0.0, 1.5, -0.1]))
        logger.info('WS defaults={}'.format(WS))

        for xw in self.config.get('xref_weights', []):
            left = xw.get('left','')
            right = xw.get('right','')
            X = np.array(xw['weights'])
            if x == left and y == right:
                WS += X
                logger.info('MATCH: {} for {}-{}'.format(X, x, y))
            elif y == left and x == right:
                WS += self._flipweights(X)
                logger.info('IMATCH: {}'.format(X))

        smap = self.smap
        # TODO: symmetrical
        WT = np.array((0.0, 0.0, 0.0, 0.0))
        WBESTMAX = np.array((0.0, 0.0, 0.0, 0.0))
        n = 0
        for sx in smap[x]:
            WBEST, _ = self._best_match_syn(sx, smap[y], scope_map)
            if WBEST is not None:
                WT += WBEST
                n += 1
                if max(abs(WBEST)) > max(abs(WBESTMAX)):
                    WBESTMAX = WBEST
        for sy in smap[y]:
            WBEST, _ = self._best_match_syn(sy, smap[x], scope_map)
            if WBEST is not None:
                WT += WBEST
                n += 1
        # average best match
        if n > 0:
            logger.info('Adding BESTMAX={}'.format(WBESTMAX))
            WS += WBESTMAX
                    
        # TODO: xref, many to many
        WS += self._graph_weights(x, y, xg)
        # TODO: include additional defined weights, eg ORDO
        logger.info('Adding WS, gw={}'.format(WS))

        # jaccard similarity
        (ss1,ss2) = xg[x][y][self.SIMSCORES]
        WS[3] += ((1-ss1) + (1-ss2)) / 2

        # reciprocal best hits are higher confidence of equiv
        rs = xg[x][y]['reciprocal_score']
        if rs == 4:
            WS[2] += 0.5
        if rs == 0:
            WS[2] -= 0.2
        
        #P = np.expit(WS)
        P = 1/(1+np.exp(-WS))
        logger.info('Final WS={}, init P={}'.format(WS, P))
        # probs should sum to 1.0
        P = P / np.sum(P)
        return P

    def _graph_weights(self, x, y, xg):
        ont = self.merged_ontology
        xancs = ont.ancestors(x)
        yancs = ont.ancestors(y)
        pfx = self._id_to_ontology(x)
        pfy = self._id_to_ontology(y)
        xns = [n for n in xg.neighbors(y) if n != x and pfx == self._id_to_ontology(n)]
        yns = [n for n in xg.neighbors(x) if n != y and pfy == self._id_to_ontology(n)]
        pweight = 1.0
        W = np.array((0,0,0,0))
        card = '11'
        if len(xns) > 0:
            card = 'm1'
            for x2 in xns:
                if x2 in xancs:
                    W[0] += pweight
                if x in ont.ancestors(x2):
                    W[1] += pweight
        if len(yns) > 0:
            if card == '11':
                card = '1m'
            else:
                card = 'mm'
            for y2 in yns:
                if y2 in yancs:
                    W[1] += pweight
                if y in ont.ancestors(y2):
                    W[0] += pweight

        logger.debug('CARD: {}/{} <-> {}/{} = {} // X={} Y={} // W={}'.format(x,pfx, y,pfy, card, xns, yns, W))
        invcard = card
        if card == '1m':
            invcard = 'm1'
        elif card == 'm1':
            invcard = '1m'
            
        CW = None
        DEFAULT_CW = None
        for cw in self.config.get('cardinality_weights', []):
            if 'prefix1' not in cw and 'prefix2' not in cw:
                if card == cw['cardinality']:
                    DEFAULT_CW = np.array(cw['weights'])
                if invcard == cw['cardinality']:
                    DEFAULT_CW = self._flipweights(np.array(cw['weights']))
            if 'prefix1'  in cw and 'prefix2' in cw:
                if pfx == cw['prefix1'] and pfy == cw['prefix2'] and card == cw['cardinality']:
                    CW = np.array(cw['weights'])
                
                if pfx == cw['prefix2'] and pfy == cw['prefix1']  and invcard == cw['cardinality']:
                    CW = self._flipweights(np.array(cw['weights']))
                    
        if CW is None:
            if DEFAULT_CW is not None:
                CW = DEFAULT_CW
            else:
                if card == '11':
                    CW = np.array((0.0, 0.0, 1.0, 0.0))
                elif card == '1m':
                    CW = np.array((0.6, 0.4, 0.0, 0.0))
                elif card == 'm1':
                    CW = np.array((0.4, 0.6, 0.0, 0.0))
                elif card == 'mm':
                    CW = np.array((0.2, 0.2, 0.0, 0.5))
        return W + CW
    
    def _flipweights(self, W):
        return np.array((W[1],W[0],W[2],W[3]))
    
    def grouped_mappings(self,id):
        """
        return all mappings for a node, grouped by ID prefix
        """
        g = self.get_xref_graph()
        m = {}
        for n in g.neighbors(id):
            [prefix, local] = n.split(':')
            if prefix not in m:
                m[prefix] = []
            m[prefix].append(n)
        return m

    def unmapped_nodes(self, xg, rs_threshold=0):
        unmapped_set = set()
        for nid in self.merged_ontology.nodes():
            if nid in xg:
                for (j,edge) in xg[nid].items():
                    rs = edge.get('reciprocal_score',0)
                    if rs < rs_threshold:
                        unmapped_set.add(nid)
            else:
                unmapped_set.add(nid)
        return unmapped_set
    
    def unmapped_dataframe(self, xg, **args):
        unodes = self.unmapped_nodes(xg, **args)
        ont = self.merged_ontology
        eg = ont.equiv_graph()
        items = []
        for n in unodes:
            mapped_equivs = ''
            if n in eg:
                equivs = set(eg.neighbors(n))
                mapped_equivs = list(equivs - unodes)
            items.append(dict(id=n,label=ont.label(n),mapped_equivs=mapped_equivs))
        df = pd.DataFrame(items, columns=['id','label', 'mapped_equivs'])
        df = df.sort_values(["id"])
        return df
            
    # scores a pairwise combination of synonyms. This will be a mix of
    #  * individual confidence in the synonyms themselves
    #  * confidence of equivalence based on scopes
    # TODO: unify this with probabilistic calculation
    def _combine_syns(self, s1,s2):
        cpred = self._combine_preds(s1.pred, s2.pred)
        s = self._pred_score(cpred)
        s *= s1.confidence * s2.confidence
        if s1.is_abbreviation() or s2.is_abbreviation():
            s *= self._get_config_val(self._id_to_ontology(s1.class_id), 'abbreviation_confidence', 0.5)
            s *= self._get_config_val(self._id_to_ontology(s1.class_id), 'abbreviation_confidence', 0.5)
        logger.debug("COMBINED: {} + {} = {}/{}".format(s1,s2,cpred,s))
        return round(s)
    
    def _rollup(self, p):
        if p == 'label':
            return LABEL_OR_EXACT
        if p == 'hasExactSynonym':
            return LABEL_OR_EXACT
        return p
    
    def _combine_preds(self, p1, p2):
        if p1 == p2:
            return p1
        if self._rollup(p1) == self._rollup(p2):
            return self._rollup(p1)
        return p1 + p2

    ## TODO: allow this to be weighted by ontology
    def _pred_score(self,p):
        if p == 'label':
            return 100
        if p == LABEL_OR_EXACT:
            return 90
        if p == 'hasExactSynonym':
            return 90
        return 50

    def _in_clique(self, x, cliques):
        for s in cliques:
            if x in s:
                return s
        return set()
    
    def as_dataframe(self, xg):
        cliques = self.cliques(xg)
        ont = self.merged_ontology
        items = []
        for (x,y,d) in xg.edges(data=True):
            # xg is a non-directional Graph object.
            # to get a deterministic ordering we use the idpair key
            (x,y) = d['idpair']
            (s1,s2)=d['syns']
            (ss1,ss2)=d['simscores']
            clique = self._in_clique(x, cliques)
            #ancs = nx.ancestors(g,x)
            left_label = ont.label(x)
            right_label = ont.label(y)
            if ont.is_obsolete(x) and not left_label.startwith('obsolete'):
                left_label = "obsolete " + left_label
            if ont.is_obsolete(y) and not right_label.startwith('obsolete'):
                right_label = "obsolete " + right_label

            P = self.weighted_axioms(x,y,xg)
            item = {'left':x, 'left_label':left_label,
                    'right':y, 'right_label':right_label,
                    'score':d['score'],
                    'left_match_type': s1.pred,
                    'right_match_type': s2.pred,
                    'left_match_val': s1.val,
                    'right_match_val': s2.val,
                    'left_simscore':ss1,
                    'right_simscore':ss2,
                    'reciprocal_score':d.get('reciprocal_score',0),
                    'conditional_pr_equiv': d.get('cpr'),
                    'pr_subClassOf': P[0],
                    'pr_superClassOf': P[1],
                    'pr_equivalentTo': P[2],
                    'pr_other': P[3],
                    'left_novel': d.get('left_novel'),
                    'right_novel': d.get('right_novel'),
                    'left_consistent': d.get('left_consistent'),
                    'right_consistent': d.get('right_consistent'),
                    'equiv_clique_size': len(clique)}
            
            items.append(item)

        ix = ['left', 'left_label', 'right', 'right_label',
              'left_match_type', 'right_match_type',
              'left_match_val', 'right_match_val', 
              'score', 'left_simscore', 'right_simscore', 'reciprocal_score',
              'conditional_pr_equiv',
              'pr_subClassOf', 'pr_superClassOf', 'pr_equivalentTo', 'pr_other',
              'left_novel',
              'right_novel',
              'left_consistent',
              'right_consistent',
              'equiv_clique_size']
        df = pd.DataFrame(items, columns=ix)
        df = df.sort_values(["left","score","right"])
        return df
    
    def cliques(self, xg):
        """
        Return all equivalence set cliques, assuming each edge in the xref graph is treated as equivalent,
        and all edges in ontology are subClassOf

        Arguments
        ---------
        xg : Graph
            an xref graph

        Returns
        -------
        list of sets
        """
        g = nx.DiGraph()
        for (x,y) in self.merged_ontology.get_graph().edges():
            g.add_edge(x,y)
        for (x,y) in xg.edges():
            g.add_edge(x,y)
            g.add_edge(y,x)
        return list(strongly_connected_components(g))
        
            
        

### MARSHMALLOW SCHEMAS
                        
class ScopeWeightMapSchema(Schema):
    """
    Maps scope predicates (label, hasExactSynonym etc) to weights (0<=1.0).

    Typically labels and exact matches have higher weight, although this
    may vary with ontology
    """
    label = fields.Float(default=1.0, description="weight of label matches")
    hasExactSynonym = fields.Float(default=0.9, description="weight of exact matches")
    hasRelatedSynonym = fields.Float(default=0.0, description="weight of related matches")
    hasBroadSynonym = fields.Float(default=-0.2, description="weight of broad matches")
    hasNarrowSynonym = fields.Float(default=-0.2, description="weight of narrow matches")
    other = fields.Float(default=-0.5, description="weight of other kinds of matches")

class OntologyConfigurationSchema(Schema):
    """
    configuration that is specific to an ontology
    """
    prefix = fields.String(description="prefix of IDs in ontology, e.g. UBERON")
    scope_weight_map = fields.Nested(ScopeWeightMapSchema(), description="local scope-weight map")
    normalized_form_confidence = fields.Float(description="confidence of a synonym value derived via normalization (e.g. canonical ordering of tokens)")
    abbreviation_confidence = fields.Float(default=0.5, description="confidence of an abbreviation")
    
class CardinalityWeights(Schema):
    """
    Weights for different cardinality combinations, 
    """
    prefix1 = fields.String(description="prefix of IDs in ontology, e.g. MA")
    prefix2 = fields.String(description="prefix of IDs in ontology, e.g. ZFA")
    cardinality = fields.String(description="One of 11, 1m, m1 or mm")
    weights = fields.List(fields.Float(), description="Sub/Sup/Eq/Other")

class MatchWeights(Schema):
    """
    Default weights for a pair of ontologies
    """
    prefix1 = fields.String(description="prefix of IDs in ontology, e.g. MA")
    prefix2 = fields.String(description="prefix of IDs in ontology, e.g. ZFA")
    weights = fields.List(fields.Float(), description="Sub/Sup/Eq/Other")
    
class XrefWeights(Schema):
    """
    Default weights for a pair of classes
    """
    left = fields.String(description="ID of first class")
    right = fields.String(description="ID of second class")
    weights = fields.List(fields.Float(), description="Sub/Sup/Eq/Other")
    
class LexicalMapConfigSchema(Schema):
    """
    global configuration
    """
    scope_weight_map = fields.Nested(ScopeWeightMapSchema(), description="global scope-weight map. May be overridden by ontologies")
    ontology_configurations = fields.List(fields.Nested(OntologyConfigurationSchema()), description="configurations that are specific to an ontology")
    normalized_form_confidence = fields.Float(default=0.8, description="confidence of a synonym value derived via normalization (e.g. canonical ordering of tokens)")
    abbreviation_confidence = fields.Float(default=0.5, description="confidence of an abbreviation")
    match_weights = fields.List(fields.Nested(MatchWeights()))
    cardinality_weights = fields.List(fields.Nested(CardinalityWeights()))
    xref_weights = fields.List(fields.Nested(XrefWeights()))



#!/usr/bin/env python

"""
Command line wrapper to gafparser.py

Type:

    ogr-parse-assocs -h

For instructions

Examples:

```
```

"""

import argparse
import networkx as nx
from networkx.algorithms.dag import ancestors, descendants
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory
from ontobio.io import entityparser
from ontobio.io.gafparser import GafParser
from ontobio.io.gpadparser import GpadParser
from ontobio.io.hpoaparser import HpoaParser
from ontobio.io.assocwriter import GafWriter, GpadWriter
from ontobio.io import assocparser
from ontobio.slimmer import get_minimal_subgraph
import json
import logging

def main():
    """
    Wrapper for Assoc Parsing
    """
    parser = argparse.ArgumentParser(description='Wrapper for obographs assocmodel library'
                                                 """
                                                 By default, ontologies and assocs are cached locally and synced from a remote sparql endpoint
                                                 """,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--resource', type=str, required=False,
                        help='Name of ontology')
    parser.add_argument('-f', '--file', type=str, required=False,
                        help='Name of input file for associations - currently GAF is assumed')
    parser.add_argument('-F', '--format', type=str, required=False,
                        help='Format of assoc file. One of GAF, GPAD or HPOA')
    parser.add_argument('-o', '--outfile', type=str, required=False,
                        help='Path to output file')
    parser.add_argument("--report-md", type=str, required=False, dest="report_md",
                        help="Path to report markdown file")
    parser.add_argument("--report-json", type=str, required=False, dest="report_json",
                        help="Path to report JSON file")
    parser.add_argument('-t', '--to', type=str, required=False,
                        help='Output to (tree, dot, ...)')
    parser.add_argument("--filter-out", nargs="+", required=False, default=[], metavar="EVIDENCE",
                        help="List of any evidence codes to filter out of the GAF. E.G. --filter-out IEA IMP")
    parser.add_argument("--filtered-file", required=False, default=None, metavar="FILTERED_FILE",
                        help="File to write the filtered out evidence GAF to")
    parser.add_argument('-T', '--taxon', nargs='*', required=False,
                        help='valid taxon (NCBITaxon ID) - validate against this')
    parser.add_argument('--subject_prefix', nargs='*', required=False,
                        help='E.g PomBase - validate against this')
    parser.add_argument('--object_prefix', nargs='*', required=False,
                        help='E.g GO - validate against this')
    parser.add_argument('-v', '--verbosity', default=0, action='count',
                        help='Increase output verbosity')


    subparsers = parser.add_subparsers(dest='subcommand', help='sub-command help')

    parser_n = subparsers.add_parser('validate', help='Validate associations')
    parser_n.set_defaults(function=validate_assocs)

    parser_n = subparsers.add_parser('filter', help='Filter associations')
    parser_n.set_defaults(function=filter_assocs)

    parser_n = subparsers.add_parser('convert', help='Convert associations')
    parser_n.set_defaults(function=convert_assocs)
    parser_n.add_argument('-t', '--to', type=str, required=True,
                          help='Format to convert to')

    parser_n = subparsers.add_parser('map2slim', help='Map to a subset/slim')
    parser_n.set_defaults(function=map2slim)
    parser_n.add_argument('-p', '--properties', nargs='*', type=str, default=['subClassOf', 'BFO:0000050'],
                          help='Properties')
    parser_n.add_argument('-s', '--subset', type=str, required=True,
                          help='subset (e.g. map2slim)')

    args = parser.parse_args()

    if args.verbosity >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    logging.info("Welcome!")

    handle = args.resource

    # Ontology Factory
    ofactory = OntologyFactory()
    logging.info("Creating ont object from: {} {}".format(handle, ofactory))
    ont = ofactory.create(handle)
    logging.info("ont: {}".format(ont))

    func = args.function

    # Upper case all evidence codes
    args.filter_out = [code.upper() for code in args.filter_out]

    # set configuration
    filtered_evidence_file = open(args.filtered_file, "w") if args.filtered_file else None
    config = assocparser.AssocParserConfig(
        valid_taxa=args.taxon,
        ontology=ont,
        class_idspaces=args.object_prefix,
        entity_idspaces=args.subject_prefix,
        filter_out_evidence=args.filter_out,
        filtered_evidence_file=filtered_evidence_file
    )
    p = None
    fmt = None
    if args.format is None:
        fmt = 'gaf'
    else:
        fmt = args.format.lower()

    # TODO: use a factory
    if fmt == 'gaf':
        p = GafParser(dataset=args.file)
    elif fmt == 'gpad':
        p = GpadParser()
    elif fmt == 'hpoa':
        p = HpoaParser()
    elif fmt == "gpi":
        p = entityparser.GpiParser()
        func = validate_entity

    p.config = config

    outfh = None
    if args.outfile is not None:
        two_mb = 2097152
        outfh = open(args.outfile, "w", buffering=two_mb)
    func(ont, args.file, outfh, p, args)
    if filtered_evidence_file:
        filtered_evidence_file.close()

    if outfh is not None:
        outfh.close()

    if args.report_md is not None:
        report_md = open(args.report_md, "w")
        report_md.write(p.report.to_markdown())
        report_md.close()
    if args.report_json is not None:
        report_json = open(args.report_json, "w")
        report_json.write(json.dumps(p.report.to_report_json(), indent=4))
        report_json.close()
    if not (args.report_md or args.report_json):
        print(p.report.to_markdown())

def filter_assocs(ont, file, outfile, p, args):
    p.generate_associations(open(file, "r"), outfile)

def validate_assocs(ont, file, outfile, p, args):
    gafwriter = GafWriter(file=outfile)

    with open(file) as gafsource:
        associations = p.association_generator(file=gafsource)
        for assoc in associations:
            gafwriter.write_assoc(assoc)

def validate_entity(ont, file, outfile, p, args):
    p.parse(open(file, "r"), outfile)

def convert_assocs(ont, file, outfile, p, args):
    assocs = p.parse(open(file, "r"), None)
    write_assocs(assocs, outfile, args)

def write_assocs(assocs, outfile, args):
    w = GpadWriter()
    fmt = args.to
    if fmt is None or fmt == 'gaf':
        w = GafWriter()
    elif fmt == 'gpad':
        w = GpadWriter()
    else:
        raise ValueError("Not supported: {}".format(fmt))
    w.file = outfile
    w.write(assocs)

def map2slim(ont, file, outfile, p, args):
    logging.info("Mapping to {}".format(args.subset))
    assocs = p.map_to_subset(open(file, "r"),
                             ontology=ont, outfile=outfile, subset=args.subset, relations=args.properties)
    #write_assocs(assocs, outfile, args)

def _default(something, default):
    if something is None:
        return default
    else:
        return something


if __name__ == "__main__":
    main()



from ontobio.io import assocwriter
from ontobio.io import gafparser
import json
import io

def test_empty_extension_expression():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    extension = writer._extension_expression({})
    assert extension == ""

def test_single_entry_extension():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    expression = {
        "union_of": [
            {
                "intersection_of": [
                    {
                        "property": "foo",
                        "filler": "X:1"
                    }
                ]
            }
        ]
    }

    extension = writer._extension_expression(expression)
    assert "foo(X:1)" == extension

def test_unions_extension():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    expression = {
        "union_of": [
            {
                "intersection_of": [
                    {
                        "property": "foo",
                        "filler": "X:1"
                    }
                ]
            },
            {
                "intersection_of": [
                    {
                        "property": "bar",
                        "filler": "Y:1"
                    }
                ]
            }
        ]
    }

    extension = writer._extension_expression(expression)
    assert "foo(X:1)|bar(Y:1)" == extension

def test_intersection_extensions():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    expression = {
        "union_of": [
            {
                "intersection_of": [
                    {
                        "property": "foo",
                        "filler": "X:1"
                    },
                    {
                        "property": "foo",
                        "filler": "X:2"
                    }
                ]
            },
            {
                "intersection_of": [
                    {
                        "property": "bar",
                        "filler": "Y:1"
                    }
                ]
            }
        ]
    }

    extension = writer._extension_expression(expression)
    assert "foo(X:1),foo(X:2)|bar(Y:1)" == extension

def test_gaf_writer():
    association = {
        "subject": {
            "id": "PomBase:SPAC25B8.17",
            "label": "ypf1",
            "type": "protein",
            "fullname": "intramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)",
            "synonyms": [
                "ppp81"
            ],
            "taxon": {
                "id": "NCBITaxon:4896"
            }
        },
        "object": {
            "id": "GO:0000006",
            "taxon": "NCBITaxon:4896"
        },
        "negated": False,
        "qualifiers": [],
        "aspect": "C",
        "relation": {
            "id": "part_of"
        },
        "interacting_taxon": "NCBITaxon:555",
        "evidence": {
            "type": "ISO",
            "has_supporting_reference": [
                "GO_REF:0000024"
            ],
            "with_support_from": [
                "SGD:S000001583"
            ]
        },
        "provided_by": "PomBase",
        "date": "20150305",
        "subject_extensions": [
            {
                "property": "isoform",
                "filler": "UniProtKB:P12345"
            }
        ],
        "object_extensions": {
            "union_of": [
                {
                    "intersection_of": [
                        {
                            "property": "foo",
                            "filler": "X:1"
                        }
                    ]
                }
            ]
        }
    }
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    expected = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896|taxon:555\t20150305\tPomBase\tfoo(X:1)\tUniProtKB:P12345"
    writer.write_assoc(association)
    gaf = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert expected == gaf

def test_full_taxon_field_single_taxon():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    taxon_field = writer._full_taxon_field("taxon:12345", None)
    assert "taxon:12345" == taxon_field

def test_full_taxon_field_interacting():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    taxon_field = writer._full_taxon_field("taxon:12345", "taxon:6789")
    assert "taxon:12345|taxon:6789" == taxon_field

def test_full_taxon_empty_string_interacting_taxon():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    taxon_field = writer._full_taxon_field("taxon:12345", "")
    assert "taxon:12345" == taxon_field

def test_roundtrip():
    """
    Start with a line, parse it, then write it. The beginning line should be the same as what was written.
    """
    line = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:999|taxon:888\t20150305\tPomBase\tfoo(X:1)\tUniProtKB:P12345"
    parser = gafparser.GafParser()
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)
    assoc_dict = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc_dict)
    gaf = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert line == gaf

    # Single taxon
    line = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:1111\t20150305\tPomBase\tfoo(X:1)\tUniProtKB:P12345"
    parser = gafparser.GafParser()
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)
    assoc_dict = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc_dict)
    gaf = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert line == gaf



import logging
import requests

import rdflib
from rdflib import Namespace
from rdflib.namespace import RDF
from rdflib.namespace import SKOS
from prefixcommons.curie_util import contract_uri

from ontobio.ontol import Ontology, Synonym

# TODO: make configurable
GEMET = Namespace('http://www.eionet.europa.eu/gemet/2004/06/gemet-schema.rdf#')

logger = logging.getLogger(__name__)


class Skos(object):
    """
    SKOS is an RDF data model for representing thesauri and terminologies.

    See https://www.w3.org/TR/skos-primer/ for more details
    """

    def __init__(self, prefixmap=None, lang='en'):
        self.prefixmap = prefixmap if prefixmap is not None else {}
        self.lang = lang
        self.context = None

    def _uri2id(self, uri):
        s = "{:s}".format(str(uri))
        for prefix,uribase in self.prefixmap.items():
            if (s.startswith(uribase)):
                s = s.replace(uribase,prefix+":")
                return s
        curies = contract_uri(uri)
        if len(curies) > 0:
            return curies[0]
        return s

    def process_file(self,filename=None, format=None):
        """
        Parse a file into an ontology object, using rdflib
        """
        rdfgraph = rdflib.Graph()
        if format is None:
            if filename.endswith(".ttl"):
                format='turtle'
            elif filename.endswith(".rdf"):
                format='xml'
        rdfgraph.parse(filename, format=format)
        return self.process_rdfgraph(rdfgraph)
        
    def process_rdfgraph(self, rg, ont=None):
        """
        Transform a skos terminology expressed in an rdf graph into an Ontology object

        Arguments
        ---------
        rg: rdflib.Graph
            graph object

        Returns
        -------
        Ontology
        """
        # TODO: ontology metadata
        if ont is None:
            ont = Ontology()
            subjs = list(rg.subjects(RDF.type, SKOS.ConceptScheme))
            if len(subjs) == 0:
                logger.warning("No ConceptScheme")
            else:
                ont.id = self._uri2id(subjs[0])
            
        subset_map = {}
        for concept in rg.subjects(RDF.type, SKOS.Concept):
            for s in self._get_schemes(rg, concept):
                subset_map[self._uri2id(s)] = s
                
        for concept in sorted(list(rg.subjects(RDF.type, SKOS.Concept))):
            concept_uri = str(concept)
            id=self._uri2id(concept)
            logger.info("ADDING: {}".format(id))
            ont.add_node(id, self._get_label(rg,concept))
                    
            for defn in rg.objects(concept, SKOS.definition):
                if (defn.language == self.lang):
                    td = TextDefinition(id, escape_value(defn.value))
                    ont.add_text_definition(td)
                    
            for s in rg.objects(concept, SKOS.broader):
                ont.add_parent(id, self._uri2id(s))
                
            for s in rg.objects(concept, SKOS.related):
                ont.add_parent(id, self._uri2id(s), self._uri2id(SKOS.related))
                
            for m in rg.objects(concept, SKOS.exactMatch):
                ont.add_xref(id, self._uri2id(m))
                
            for m in rg.objects(concept, SKOS.altLabel):
                syn = Synonym(id, val=self._uri2id(m))
                ont.add_synonym(syn)
                
            for s in self._get_schemes(rg,concept):
                ont.add_to_subset(id, self._uri2id(s))
                
        return ont
    
    def _get_schemes(self, rg, concept):
        schemes = set(rg.objects(concept, SKOS.inScheme))
        schemes.update(rg.objects(concept, GEMET.group))
        return schemes
    
    def _get_label(self, rg,concept):
        labels = sorted(rg.preferredLabel(concept, lang=self.lang))
        if len(labels) == 0:
            return None
        if len(labels) > 1:
            logger.warning(">1 label for {} : {}".format(concept, labels))
        return labels[0][1].value
    
    



from ontobio.io.gpadparser import GpadParser
from ontobio.io import assocparser

import yaml

POMBASE = "tests/resources/truncated-pombase.gpad"

def test_skim():
    p = GpadParser()
    results = p.skim(open(POMBASE,"r"))
    print(str(results))


def test_parse():
    p = GpadParser(config=assocparser.AssocParserConfig(group_metadata=yaml.load(open("tests/resources/mgi.dataset.yaml"), Loader=yaml.FullLoader)))
    test_gpad_file = "tests/resources/mgi.test.gpad"
    results = p.parse(open(test_gpad_file, "r"))
    print(p.report.to_markdown())



import typing
import enum
import io
import collections
import click
import json

from dataclasses import dataclass
from typing import List, Dict, TypeVar, Union, Generic, Optional

from ontobio.validation import metadata
from ontobio.io import assocparser
from ontobio.io import gafparser
from ontobio.io import assocwriter

"""
This module is for centralizing logic related to validating the example data in
GO Rules. Optional data can be included in rules that show exmaples of failing,
passing, or repair type rules for incoming data (either GAF, GPAD, or RDF).

This will first just support GAF for the first pass in all liklihood.

Relavent schema:
```
"examples":
  type: map
    required: false
    mapping:
      "pass":
        type: seq
        required: false
        sequence:
          - type: map
            mapping:
              "comment":
                type: str
                required: true
              "format":
                type: str
                required: true
                enum: ["rdf", "gaf", "gpad"]
              "input":
                type: str
                required: true
      "fail":
        type: seq
        required: false
        sequence:
          - type: map
            mapping:
              "comment":
                type: str
                required: true
              "format":
                type: str
                required: true
                enum: ["rdf", "gaf", "gpad"]
              "input":
                type: str
                required: true
      "repair":
        type: seq
        required: false
        sequence:
          - type: map
            mapping:
              "comment":
                type: str
                required: true
              "format":
                type: str
                required: true
                enum: ["rdf", "gaf", "gpad"]
              "input":
                type: str
                required: true
              "output":
                type: str
                required: true
```

"""
FormatType = enum.Enum("FormatType", ["RDF", "GAF", "GPAD"])
ExampleType = enum.Enum("ExampleType", {"REPAIR": "repair", "FAIL": "fail", "PASS": "fail"})

def format_from_string(format: str) -> Optional[FormatType]:
    if format == "rdf":
        return FormatType.RDF

    if format == "gaf":
        return FormatType.GAF

    if format == "gpad":
        return FormatType.GPAD

    return None

@dataclass
class RuleExample:
    rule_id: str
    example_type: ExampleType
    input: str
    format: FormatType
    expected: Union[str, bool]

    @classmethod
    def example_from_json(RuleExample, rule_json: Dict) -> List:
        """
        This constructs the set of examples to be run from a single GO rule.
        """
        # Returns List of RuleExample
        if "examples" not in rule_json:
            # Bail if we don't have any examples
            return []

        fail_examples = rule_json["examples"].get("fail", [])
        pass_examples = rule_json["examples"].get("pass", [])
        repair_examples = rule_json["examples"].get("repair", [])

        built_examples = [] # type: List[RuleExample]
        ruleid = rule_json["id"].lower().replace(":", "-")
        for ex in fail_examples:
            f = format_from_string(ex["format"]) # type: Optional[FormatType]
            # Expected is False since these are "fail" examples
            built_examples.append(RuleExample(ruleid, ExampleType.FAIL, ex["input"], f, False))

        for ex in pass_examples:
            f = format_from_string(ex["format"]) # type: Optional[FormatType]
            # Expected is True since these are "pass" examples
            built_examples.append(RuleExample(ruleid, ExampleType.PASS, ex["input"], f, True))

        for ex in repair_examples:
            f = format_from_string(ex["format"]) # type: Optional[FormatType]
            # Expected will come from the `output` field
            built_examples.append(RuleExample(ruleid, ExampleType.REPAIR, ex["input"], f, ex["output"]))

        return built_examples

@dataclass
class ValidationResult:
    example: RuleExample
    actual: Union[str, bool]
    success: bool
    reason: str

    def to_json(self) -> Dict:
        return {
            "rule": self.example.rule_id,
            "type": self.example.example_type.value,
            "format": self.example.format.value,
            "input": self.example.input,
            "expected": self.example.expected,
            "actual": self.actual,
            "success": self.success,
            "reason": self.reason
        }

Parsed = collections.namedtuple("Parsed", ["report", "output", "expected"])

#==============================================================================

def normalize_tsv_row(size: int, tsv: str) -> str:
    columns = tsv.split("\t")
    if len(columns) < size:
        columns += [""] * (size - len(columns))
    elif len(columns) > size:
        columns = columns[0:size]

    return "\t".join(columns)


def validate_all_examples(examples: List[RuleExample], config=None) -> List[ValidationResult]:
    results = []
    for ex in examples:
        r = validate_example(ex, config=config)
        results.append(r)

    return results


def validate_example(example: RuleExample, config=None) -> ValidationResult:
    """
    1. Create parser based on `format`
    2. Run input into parser/validator
    3. In parsed rule results, find the results for the rule given in the example
    4. Decide on what the `output` is.
    5. Validate output against the example `expected` to decide on `success`
    6. Consolodate and return `ValidationResult`
    """
    parser = create_base_parser(example.format)
    parsed = validate_input(example, parser, config=config)
    # click.echo(parsed)
    success = example_success(example, parsed)

    actual = len(parsed.report) == 0 if example.example_type in [ExampleType.FAIL, ExampleType.PASS] else parsed.output
    reason = "Valid"
    if not success:
        if example.example_type in [ ExampleType.PASS or ExampleType.FAIL ]:
            reason = "Input was expected to {passfail} {ruleid}, but it did not: {message}".format(passfail=example.example_type.value, ruleid=example.rule_id,
                message="; ".join([m["message"] for m in parsed.report]))
        else:
            reason = "Repair found `{}`, but expected `{}`".format(actual, example.expected)

    result = ValidationResult(example, actual, success, reason)
    return result


def create_base_parser(format: FormatType) -> Optional[assocparser.AssocParser]:
    """
    Make an unconfigured parser based on the format. Only GAF is supported currently.
    """
    parser = None
    if format == FormatType.GAF:
        parser = gafparser.GafParser(config=assocparser.AssocParserConfig())
    else:
        parser = None

    return parser


def validate_input(example: RuleExample, parser: assocparser.AssocParser, config=None) -> Parsed:
    if config:
        parser.config = config

    out = []
    writer = assocwriter.GafWriter(file=io.StringIO())

    assocs_gen = parser.association_generator(file=io.StringIO(example.input), skipheader=True)
    for assoc in assocs_gen:
        out.append(writer.tsv_as_string(writer.as_tsv(assoc)))

    rule_messages = parser.report.reporter.messages.get(example.rule_id, [])
    rule_messages.extend(parser.report.reporter.messages.get("gorule-0000001", []))

    # We have to also parse the expected result if we are in a repair to normalize all the data
    expected_out = []
    if example.example_type == ExampleType.REPAIR:
        expected_parsed_gen = create_base_parser(example.format).association_generator(file=io.StringIO(example.expected), skipheader=True)
        expected_writer = assocwriter.GafWriter(file=io.StringIO())
        for assoc in expected_parsed_gen:
            expected_out.append(expected_writer.tsv_as_string(expected_writer.as_tsv(assoc)))

    # We only collect the messages from *our* rule we're in
    return Parsed(report=rule_messages, output="\n".join(out), expected="\n".join(expected_out))


def example_success(example: RuleExample, parsed_input: Parsed) -> bool:
    """
    Decide if the example was a success. Given the example and the result of parsing
    and validation, this will return True if we expected the example to fail validation/repair
    the rule and it did so or if the example is expected to pass the rule and it did so.
    This returns False if validation produces something we did not expect.

    Additionally, examples will not succeed if we aren't testing rule 1, but still fail rule 1
    """
    success = False
    if example.rule_id != "gorule-0000001" and (1 in [message["rule"] for message in parsed_input.report]):
        # If we find a gorule-0000001 report in a non gorule-0000001 example, than we auto-fail success
        return False

    if example.example_type == ExampleType.REPAIR:
        success = parsed_input.output == parsed_input.expected
    elif example.example_type in [ ExampleType.FAIL, ExampleType.PASS ]:
        # The rule was passed if there were no messages from that rule
        passed_rule = len(parsed_input.report) == 0
        # We have a successful example if the passing above was what we expected give the example
        success = passed_rule == example.expected

    return success

def validation_report(all_results: List[ValidationResult]) -> Dict:
    """
    {
        "gorule-00000008": {
            "results": [
                {
                    "rule": "gorule-00000008"
                    "type": "fail|pass|repair",
                    "format": "gaf|gpad|rdf",
                    "input": "string",
                    "expected": "string|bool",
                    "actual": "string|bool",
                    "success": "bool",
                    "reason": "string"
                }, "..."
            ]
        }
    }
    """
    report_dict = dict()
    for r in all_results:
        if r.example.rule_id not in report_dict:
            report_dict[r.example.rule_id] = []

        r_out = r.to_json()
        report_dict[r.example.rule_id].append(r_out)

    return report_dict



from enum import Enum


class PairwiseDist(Enum):
    EUCLIDEAN   = 'euclidean'
    JIN_CONRATH = 'jin_conrath'


class PairwiseSim(Enum):
    GEOMETRIC = 'geometric'  # geometric mean of ic and jaccard (phenodigm)
    IC        = 'ic'
    JACCARD   = 'jaccard'


class MatrixMetric(Enum):
    MAX = 'max'  # Max
    AVG = 'avg'  # Average
    BMA = 'bma'  # Best Match Average


class SimAlgorithm(Enum):
    JACCARD = 'jaccard'
    PHENODIGM = 'phenodigm'
    RESNIK = 'resnik'
    SYMMETRIC_RESNIK = 'symmetric_resnik'
    SIM_GIC = 'simGIC'
    COSINE = 'cosine'
    NAIVE_BAYES_TWO_STATE = 'naive-bayes-fixed-weight-two-state'
    NAIVE_BAYES_TWO_STATE_NO_BLANKET = 'naive-bayes-fixed-weight-two-state-NOBLANKET'
    MAX_INFORMATION = 'max-information'
    BAYES_NETWORK = 'bayesian-network'
    GRID = 'grid'
    NAIVE_BAYES_THREE_STATE = 'naive-bayes-fixed-weight-three-state'
    BAYES_VARIABLE = 'bayes-variable'
    GRID_NEGATED = 'grid-negated'


from marshmallow import Schema, fields, pprint

class NamedObjectSchema(Schema):
    id = fields.Str(readOnly=True, description='ID or CURIE e.g. MGI:1201606')
    label = fields.Str(readOnly=True, description='RDFS Label')
    description = fields.Str(readOnly=True, description='Descriptive text for the entity. For ontology classes, this will be a definition.')
    categories = fields.List(fields.String(readOnly=True, description='Type of object (inferred)'))
    types = fields.List(fields.String(readOnly=True, description='Type of object (direct)'))
    synonyms = fields.List(fields.Nested(SynonymPropertyValueSchema), description='list of synonyms or alternate labels')
    deprecated = fields.Boolean(description='True if the node is deprecated/obsoleted.')
    replaced_by = fields.List(fields.String(readOnly=True, description='Direct 1:1 replacement (if named object is obsoleted)'))
    consider = fields.List(fields.String(readOnly=True, description='Potential replacement object (if named object is obsoleted)'))

class EntityReferenceSchema(Schema):
    id = fields.Str(readOnly=True, description='ID or CURIE e.g. MGI:1201606')
    label = fields.Str(readOnly=True, description='RDFS Label')
    categories = fields.List(fields.String(readOnly=True, description='Type of object'))

class RelationSchema(NamedObjectSchema):
    pass

class PublicationSchema(NamedObjectSchema):
    pass

class TaxonSchema(Schema):
    id = fields.Str(readOnly=True, description='CURIE ID, e.g. NCBITaxon:9606')
    label = fields.Str(readOnly=True, description='RDFS Label')

class BioObjectSchema(NamedObjectSchema):
    taxon = fields.Nested(TaxonSchema, description='Taxon to which the object belongs')
    xrefs = fields.List(fields.String, description='Database cross-references. These are usually CURIEs, but may also be URLs. E.g. ENSEMBL:ENSG00000099940 ')

class AnnotationExtensionSchema(Schema):
    relation_chain = fields.List(fields.Nested(RelationSchema), description='Relationship type. If more than one value, interpreted as chain')
    filler = fields.Nested(NamedObjectSchema, description='Extension interpreted as OWL expression (r1 some r2 some .. some filler).')

class AssociationSchema(Schema):
    id = fields.Str(readOnly=True, description='Association/annotation unique ID')
    type = fields.Str(readOnly=True, description='Type of association, e.g. gene-phenotype')
    subject = fields.Nested(BioObjectSchema, description='Subject of association (what it is about), e.g. ClinVar:nnn, MGI:1201606')
    subject_extension = fields.List(fields.Nested(AnnotationExtensionSchema, description='Additional properties of the subject in the context of this association.'))
    object = fields.Nested(BioObjectSchema, description='Object (sensu RDF), aka target, e.g. HP:0000448, MP:0002109, DOID:14330')
    object_extension = fields.List(fields.Nested(AnnotationExtensionSchema, description='Additional properties of the object in the context of this association. See http://www.biomedcentral.com/1471-2105/15/155'))
    relation = fields.Nested(RelationSchema, description='Relationship type connecting subject and object')
    qualifiers = fields.List(fields.Nested(AssociationPropertyValueSchema, description='Qualifier on the association'))
    evidence_types = fields.List(fields.Nested(NamedObjectSchema), description='Evidence types (ECO classes) extracted from evidence graph')
    provided_by = fields.List(fields.String, description='Provider of association, e.g. Orphanet, ClinVar')
    publications = fields.List(fields.Nested(PublicationSchema), description='Publications supporting association, extracted from evidence graph')



#!/usr/bin/env python

"""
Command line wrapper to ontobio lexmap library

Type:

    ogr -h

For instructions

"""

import argparse
import networkx as nx
import pandas as pd
import numpy as np
from networkx.algorithms.dag import ancestors, descendants
from ontobio.ontol import Ontology
from ontobio.ontol_factory import OntologyFactory
from ontobio.io.ontol_renderers import GraphRenderer
from ontobio.slimmer import get_minimal_subgraph
from prefixcommons.curie_util import contract_uri, expand_uri
from ontobio.lexmap import LexicalMapEngine
import logging
import yaml
import json

def main():
    """
    Wrapper for OGR
    """
    parser = argparse.ArgumentParser(description='Wrapper for ontobio lexical mapping'
                                                 """
                                                 Lexically maps one or more ontologies. Ontologies can be local or remote,
                                                 any input handle can be specified, see docs for more details on handles.

                                                 If multiple ontologies are specified, then each ontology in the list is compared against the first one.

                                                 If a simgle ontology is specified, then all pairs in that ontology will be compared
                                                 
                                                 Output format to be documented - see lexmap.py for the various scoring attributes for now.
                                                 """,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-o', '--outfile', type=str, nargs='*', default=[], required=False,
                        help='Path to output file')
    parser.add_argument('-t', '--to', type=str, required=False, default='tsv',
                        help='Output to (tree, dot, ...)')
    parser.add_argument('-l', '--labels', type=str,
                        help='If set, then include node labels in results. DEPRECATED')
    parser.add_argument('-s', '--scoring', default='sim', type=str,
                        help='Score weighting scheme. Default=sim')
    parser.add_argument('-P', '--prefix', type=str, required=False,
                        help='Prefix to constrain traversal on, e.g. PATO, ENVO')
    parser.add_argument('-c', '--config', type=str, required=False,
                        help='lexmap configuration file (yaml). See schema for details')
    parser.add_argument('-X', '--xref_weights', type=str, required=False,
                        help='csv of curated per-xref weights')
    parser.add_argument('-e', '--eval_xrefs', default=False, action='store_true', dest='eval_xrefs',
                        help='only compare xrefs')
    parser.add_argument('-u', '--unmapped', type=str, required=False,
                        help='File to export unmapped nodes to')
    parser.add_argument('-A', '--all-by-all', dest='all_by_all', action='store_true',
                        help='compare all ontologies against all.')
    parser.add_argument('-v', '--verbosity', default=0, action='count',
                        help='Increase output verbosity')

    parser.add_argument('ontologies',nargs='*',
                        help='one or more ontologies to be aligned. Any input handle can be specified')

    args = parser.parse_args()

    if args.verbosity >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
        
    logging.info("Welcome!")

    factory = OntologyFactory()
    onts = [filter_by_prefix(factory.create(h)) for h in args.ontologies]

    
    config = {}
    if args.config is not None:
        f = open(args.config,'r')
        config = yaml.load(f, Loader=yaml.FullLoader)
        f.close()

    # add pre-defined weights to config
    if args.xref_weights is not None:
        if 'xref_weights' not in config:
            config['xref_weights'] = []
        xws = config['xref_weights']
        df = pd.read_csv(args.xref_weights)
        df = df.fillna(0.0)
        for _, row in df.iterrows():
            w = float(row['weight'])
            WA = np.array((0.0, 0.0, 0.0, 0.0))
            if w < 0:
                WA[2] = w
                WA[3] = abs(w)
            else:
                WA[2] = w
                WA[3] = -w
            xws.append({'left':row['left'],
                        'right':row['right'],
                        'weights':WA})
        
    logging.info("ALL: {}".format(args.all_by_all))
    
    lexmap = LexicalMapEngine(config=config)

    if args.eval_xrefs:
        cpairs = []
        for n in onts[0].nodes():
            for x in onts[0].xrefs(n):
                cpairs.append((n,x))
        lexmap.class_pairs = cpairs
        logging.info("Comparing {} class pairs. Sample: {}".format(len(cpairs), cpairs[0:50]))
    
    
    
    if len(onts) == 0:
        raise ValueException("must pass one or more ontologies")
    else:
        logging.info("Indexing ontologies: {}".format(onts))
        for ont in onts:
            lexmap.index_ontology(ont)
        oid0 = onts[0].id
        pairs = [(oid0,oid0)]
        if len(onts) > 1:
            if args.all_by_all:
                logging.info("All vs ALL: {}".format(onts))
                pairs = []
                for i in onts:
                    for j in onts:
                        if i.id < j.id:
                            pairs.append((i.id, j.id))
            else:
                logging.info("All vs first in list: {}".format(oid0))
                pairs = [(oid0, ont.id) for ont in onts[1:]]
        logging.info("Comparing the following pairs of ontologies: {}".format(pairs))
        lexmap.ontology_pairs = pairs
    mo = Ontology()
    mo.merge(onts)

    g = lexmap.get_xref_graph()
    
    
    if args.to == 'obo':
        write_obo(g,mo,args)
    else:
        write_tsv(lexmap,g,mo,args)

        
    if args.unmapped:
        udf = lexmap.unmapped_dataframe(g)
        udf.to_csv(args.unmapped, sep="\t", index=False)
        
def write_tsv(lexmap,g,mo,args):
    df=lexmap.as_dataframe(g)
    print(df.to_csv(sep="\t", index=False))

# TODO    
def write_json(lexmap,g,fn,args):
    nodes = []
    for x,y,d in g.edges_iter(data=True):
        nodes.append({'id':x,
                      'meta':{
                      }})
    f = open(fn,'write')
    json.dump({'graphs':{'nodes':nodes}},f)
    f.close()
    
def write_obo(g,mo,args):
    for x,y,d in g.edges_iter(data=True):
        score=str(d['score'])
        (s1,s2)=d['syns']
        print('[Term]')
        print('id: {} ! {}'.format(x,mo.label(x)))
        print('xref: {} ! {} // {} {} {}'.format(y,mo.label(y),score,s1.val,s2.val))
        print()

def in_clique(x, cliques):
    for s in cliques:
        if x in s:
            return s
    return set()

def filter_by_prefix(ont):
    if ont.handle:
        pfx = ont.handle
        nids = [n for n in ont.nodes() if ont.prefix(n).lower() == pfx]
        if len(nids) > 20:
            logging.info("filtering {} to {} nodes with prefix {}".format(ont,len(nids),pfx))
            ont = ont.subontology(nids)
    return ont

if __name__ == "__main__":
    main()
